@@ -6,6 +6,8 @@ import sklearn.svm
 import sklearn.ensemble
 import sklearn.multioutput
 import sklearn.gaussian_process
+import sklearn.model_selection
+import numpy
 import xgboost as xgb
 import diptest
 from sklearn.model_selection import train_test_split
@@ -15,6 +17,9 @@ from sklearn.metrics import (
     f1_score,
     precision_score,
     recall_score,
+    roc_auc_score,
+    accuracy_score,
+    log_loss,
 )
 
 import pickle
@@ -24,14 +29,6 @@ from pgml_extension.exceptions import PgMLException
 from pgml_extension.sql import q
 
 
-def flatten(S):
-    if S == []:
-        return S
-    if isinstance(S[0], list):
-        return flatten(S[0]) + flatten(S[1:])
-    return S[:1] + flatten(S[1:])
-
-
 class Project(object):
     """
     Use projects to refine multiple models of a particular dataset on a specific objective.
@@ -147,6 +144,20 @@ class Project(object):
             model.deploy(qualifier)
         return model
 
+    @property
+    def key_metric_name(self):
+        if self.objective == "classification":
+            return "f1"
+        elif self.objective == "regression":
+            return "r2"
+
+    @property
+    def hyperparam_score_name(self):
+        if self.objective == "classification":
+            return "f1_micro"
+        elif self.objective == "regression":
+            return "r2"
+
     @property
     def last_snapshot(self):
         return Snapshot.last_for_project_id(self.id)
@@ -185,7 +196,7 @@ class Snapshot(object):
         Args:
             relation_name (str): the name of the table or view to snapshot
             y_column_name (str): the label for training data
-            test_size (float or int, optional): If float, should be between 0.0 and 1.0 and represent the proportion of the dataset to include in the test split. If int, represents the absolute number of test samples. If None, the value is set to the complement of the train size. If train_size is also None, it will be set to 0.25.
+            test_size (float or int, optional): If float, should be between 0.0 and 1.0 and represent the proportion of the dataset to include in the test split. If int, represents the absolute number of test samples. If None, the value is set to 0.25.
             test_sampling: (str, optional): How to sample to create the test data. Defaults to "random". Valid values are ["first", "last", "random"].
         Returns:
             Snapshot: metadata instantiated from the database
@@ -202,13 +213,14 @@ class Snapshot(object):
                 1,
             )[0]
         )
-        plpy.execute(
-            f"""
+        sql = f"""
             CREATE TABLE pgml."snapshot_{snapshot.id}" AS 
             SELECT * FROM {snapshot.relation_name}
-            ORDER BY random();
         """
-        )
+        if snapshot.test_sampling == "random":
+            sql += "ORDER BY random()"
+        plpy.execute(sql)
+
         snapshot.analyze()
         snapshot.__dict__ = dict(
             plpy.execute(
@@ -266,7 +278,7 @@ class Snapshot(object):
         sample = plpy.execute(
             f"""
             SELECT * 
-            FROM pgml."snapshot_{self.id}"
+            FROM "pgml"."snapshot_{self.id}"
             LIMIT 1000
         """
         )
@@ -275,7 +287,7 @@ class Snapshot(object):
             raise PgMLException(
                 f"Relation `{self.relation_name}` contains no rows. Did you pass the correct `relation_name`?"
             )
-        
+
         for column in self.y_column_name:
             if not column in sample[0]:
                 raise PgMLException(f"Column `{column}` not found. Did you pass the correct `y_column_name`?")
@@ -294,46 +306,44 @@ class Snapshot(object):
                 self.analysis[f"{column}_dip"] = diptest.dipstat(data)
 
         self.columns = {}
-        parts = self.relation_name.split(".")
-        sql = "SELECT column_name, data_type FROM information_schema.columns "
-        if len(parts) > 1:
-            sql += f"WHERE table_name = {q(parts[1])} and table_schema = {q(parts[0])}"
-        else:
-            sql += f"WHERE table_name = {q(parts[0])}"
-        result = plpy.execute(sql)
+        result = plpy.execute(
+            f"""
+                SELECT column_name, data_type 
+                FROM information_schema.columns 
+                WHERE table_name = 'snapshot_{self.id}' and table_schema = {q("pgml")}
+            """
+        )
         for row in result:
             self.columns[row["column_name"]] = row["data_type"]
 
         self.__dict__ = dict(
             plpy.execute(
                 f"""
-            UPDATE pgml.snapshots 
-            SET columns = {q(json.dumps(self.columns))}::JSON, analysis = {q(json.dumps(self.analysis))}::JSON, updated_at = clock_timestamp()
-            WHERE id = {q(self.id)} 
-            RETURNING *
-        """,
+                    UPDATE pgml.snapshots 
+                    SET columns = {q(self.columns)}, analysis = {q(self.analysis)}, updated_at = clock_timestamp()
+                    WHERE id = {q(self.id)} 
+                    RETURNING *
+                """,
                 1,
             )[0]
         )
 
+    @property
+    def snapshot_name(self):
+        return f'"pgml"."snapshot_{self.id}"'
+
     def data(self):
         """
         Returns:
             list, list, list, list: All rows from the snapshot split into X_train, X_test, y_train, y_test sets.
         """
-        data = plpy.execute(
-            f"""
-            SELECT * 
-            FROM pgml."snapshot_{self.id}"
-        """
-        )
+        data = plpy.execute(f"SELECT * FROM {self.snapshot_name}")
 
         # Always pull the columns in the same order from the row.
         # Python dict iteration is not always in the same order (hash table).
         features = list(data[0].keys())
         for column in self.y_column_name:
             features.remove(column)
-        features.sort()
 
         # Split the label from the features
         X = []
@@ -347,7 +357,7 @@ class Snapshot(object):
             for feature in features:
                 x_.append(row[feature])
 
-            x_ = flatten(x_)  # TODO be smart about flattening X depending on algorithm
+            x_ = numpy.array(x_).flatten()  # TODO be smart about flattening X depending on algorithm
             X.append(x_)
             y.append(y_)
 
@@ -383,6 +393,56 @@ class Model(object):
         algorithm: the in memory version of the model parameters that can make predictions
     """
 
+    @classmethod
+    def algorithm_from_name_and_objective(cls, name: str, objective: str):
+        return {
+            "linear_regression": sklearn.linear_model.LinearRegression,
+            "linear_classification": sklearn.linear_model.LogisticRegression,
+            "ridge_regression": sklearn.linear_model.Ridge,
+            "ridge_classification": sklearn.linear_model.RidgeClassifier,
+            "lasso_regression": sklearn.linear_model.Lasso,
+            "elastic_net_regression": sklearn.linear_model.ElasticNet,
+            "least_angle_regression": sklearn.linear_model.Lars,
+            "lasso_least_angle_regression": sklearn.linear_model.LassoLars,
+            "orthoganl_matching_pursuit_regression": sklearn.linear_model.OrthogonalMatchingPursuit,
+            "bayesian_ridge_regression": sklearn.linear_model.BayesianRidge,
+            "automatic_relevance_determination_regression": sklearn.linear_model.ARDRegression,
+            "stochastic_gradient_descent_regression": sklearn.linear_model.SGDRegressor,
+            "stochastic_gradient_descent_classification": sklearn.linear_model.SGDClassifier,
+            "perceptron_classification": sklearn.linear_model.Perceptron,
+            "passive_aggressive_regression": sklearn.linear_model.PassiveAggressiveRegressor,
+            "passive_aggressive_classification": sklearn.linear_model.PassiveAggressiveClassifier,
+            "ransac_regression": sklearn.linear_model.RANSACRegressor,
+            "theil_sen_regression": sklearn.linear_model.TheilSenRegressor,
+            "huber_regression": sklearn.linear_model.HuberRegressor,
+            "quantile_regression": sklearn.linear_model.QuantileRegressor,
+            "kernel_ridge_regression": sklearn.kernel_ridge.KernelRidge,
+            "gaussian_process_regression": sklearn.gaussian_process.GaussianProcessRegressor,
+            "gaussian_process_classification": sklearn.gaussian_process.GaussianProcessClassifier,
+            "svm_regression": sklearn.svm.SVR,
+            "svm_classification": sklearn.svm.SVC,
+            "nu_svm_regression": sklearn.svm.NuSVR,
+            "nu_svm_classification": sklearn.svm.NuSVC,
+            "linear_svm_regression": sklearn.svm.LinearSVR,
+            "linear_svm_classification": sklearn.svm.LinearSVC,
+            "ada_boost_regression": sklearn.ensemble.AdaBoostRegressor,
+            "ada_boost_classification": sklearn.ensemble.AdaBoostClassifier,
+            "bagging_regression": sklearn.ensemble.BaggingRegressor,
+            "bagging_classification": sklearn.ensemble.BaggingClassifier,
+            "extra_trees_regression": sklearn.ensemble.ExtraTreesRegressor,
+            "extra_trees_classification": sklearn.ensemble.ExtraTreesClassifier,
+            "gradient_boosting_trees_regression": sklearn.ensemble.GradientBoostingRegressor,
+            "gradient_boosting_trees_classification": sklearn.ensemble.GradientBoostingClassifier,
+            "hist_gradient_boosting_regression": sklearn.ensemble.HistGradientBoostingRegressor,
+            "hist_gradient_boosting_classification": sklearn.ensemble.HistGradientBoostingClassifier,
+            "random_forest_regression": sklearn.ensemble.RandomForestRegressor,
+            "random_forest_classification": sklearn.ensemble.RandomForestClassifier,
+            "xgboost_regression": xgb.XGBRegressor,
+            "xgboost_classification": xgb.XGBClassifier,
+            "xgboost_random_forest_regression": xgb.XGBRFRegressor,
+            "xgboost_random_forest_classification": xgb.XGBRFClassifier,
+        }[name + "_" + objective]
+
     @classmethod
     def create(
         cls,
@@ -390,6 +450,9 @@ class Model(object):
         snapshot: Snapshot,
         algorithm_name: str,
         hyperparams: dict,
+        search: str,
+        search_params: dict,
+        search_args: dict,
     ):
         """
         Create a Model and save it to the database.
@@ -403,8 +466,8 @@ class Model(object):
         """
         result = plpy.execute(
             f"""
-            INSERT INTO pgml.models (project_id, snapshot_id, algorithm_name, hyperparams, status, created_at, updated_at) 
-            VALUES ({q(project.id)}, {q(snapshot.id)}, {q(algorithm_name)}, {q(json.dumps(hyperparams))}, 'new', clock_timestamp(), clock_timestamp()) 
+            INSERT INTO pgml.models (project_id, snapshot_id, algorithm_name, hyperparams, status, search, search_params, search_args, created_at, updated_at) 
+            VALUES ({q(project.id)}, {q(snapshot.id)}, {q(algorithm_name)}, {q(hyperparams)}, 'new', {q(search)}, {q(search_params)}, {q(search_args)}, clock_timestamp(), clock_timestamp()) 
             RETURNING *
         """
         )
@@ -463,9 +526,9 @@ class Model(object):
 
         if strategy == "best_score":
             if project.objective == "regression":
-                sql += f"{where}\nORDER BY models.metrics->>'mean_squared_error' DESC NULLS LAST"
+                sql += f"{where}\nORDER BY models.metrics->>'r2' DESC NULLS LAST"
             elif project.objective == "classification":
-                sql += f"{where}\nORDER BY models.metrics->>'f1' ASC NULLS LAST"
+                sql += f"{where}\nORDER BY models.metrics->>'f1' DESC NULLS LAST"
         elif strategy == "most_recent":
             sql += f"{where}\nORDER by models.created_at DESC"
         elif strategy == "rollback":
@@ -519,6 +582,24 @@ class Model(object):
         model.__init__()
         return model
 
+    @classmethod
+    def find_by_id(cls, pk: int):
+        """Find the model by primary key in pgml.models"""
+        result = plpy.execute(
+            f"""
+                SELECT models.*
+                FROM pgml.models
+                WHERE id = {q(pk)}
+            """
+        )
+        if len(result) == 0:
+            return None
+
+        model = Model()
+        model.__dict__ = dict(result[0])
+        model.__init__()
+        return model
+
     def __init__(self):
         self._algorithm = None
         self._project = None
@@ -527,6 +608,10 @@ class Model(object):
             self.hyperparams = json.loads(self.hyperparams)
         if "metrics" in self.__dict__ and type(self.metrics) is str:
             self.metrics = json.loads(self.metrics)
+        if "search_params" in self.__dict__ and type(self.search_params) is str:
+            self.search_params = json.loads(self.search_params)
+        if "search_args" in self.__dict__ and type(self.search_args) is str:
+            self.search_args = json.loads(self.search_args)
 
     @property
     def project(self):
@@ -554,52 +639,9 @@ class Model(object):
             if self.pickle is not None:
                 self._algorithm = pickle.loads(self.pickle)
             else:
-                self._algorithm = {
-                    "linear_regression": sklearn.linear_model.LinearRegression,
-                    "linear_classification": sklearn.linear_model.LogisticRegression,
-                    "ridge_regression": sklearn.linear_model.Ridge,
-                    "ridge_classification": sklearn.linear_model.RidgeClassifier,
-                    "lasso_regression": sklearn.linear_model.Lasso,
-                    "elastic_net_regression": sklearn.linear_model.ElasticNet,
-                    "least_angle_regression": sklearn.linear_model.Lars,
-                    "lasso_least_angle_regression": sklearn.linear_model.LassoLars,
-                    "orthoganl_matching_pursuit_regression": sklearn.linear_model.OrthogonalMatchingPursuit,
-                    "bayesian_ridge_regression": sklearn.linear_model.BayesianRidge,
-                    "automatic_relevance_determination_regression": sklearn.linear_model.ARDRegression,
-                    "stochastic_gradient_descent_regression": sklearn.linear_model.SGDRegressor,
-                    "stochastic_gradient_descent_classification": sklearn.linear_model.SGDClassifier,
-                    "perceptron_classification": sklearn.linear_model.Perceptron,
-                    "passive_aggressive_regression": sklearn.linear_model.PassiveAggressiveRegressor,
-                    "passive_aggressive_classification": sklearn.linear_model.PassiveAggressiveClassifier,
-                    "ransac_regression": sklearn.linear_model.RANSACRegressor,
-                    "theil_sen_regression": sklearn.linear_model.TheilSenRegressor,
-                    "huber_regression": sklearn.linear_model.HuberRegressor,
-                    "quantile_regression": sklearn.linear_model.QuantileRegressor,
-                    "kernel_ridge_regression": sklearn.kernel_ridge.KernelRidge,
-                    "gaussian_process_regression": sklearn.gaussian_process.GaussianProcessRegressor,
-                    "gaussian_process_classification": sklearn.gaussian_process.GaussianProcessClassifier,
-                    "svm_regression": sklearn.svm.SVR,
-                    "svm_classification": sklearn.svm.SVC,
-                    "nu_svm_regression": sklearn.svm.NuSVR,
-                    "nu_svm_classification": sklearn.svm.NuSVC,
-                    "linear_svm_regression": sklearn.svm.LinearSVR,
-                    "linear_svm_classification": sklearn.svm.LinearSVC,
-                    "ada_boost_regression": sklearn.ensemble.AdaBoostRegressor,
-                    "ada_boost_classification": sklearn.ensemble.AdaBoostClassifier,
-                    "bagging_regression": sklearn.ensemble.BaggingRegressor,
-                    "bagging_classification": sklearn.ensemble.BaggingClassifier,
-                    "extra_trees_regression": sklearn.ensemble.ExtraTreesRegressor,
-                    "extra_trees_classification": sklearn.ensemble.ExtraTreesClassifier,
-                    "gradient_boosting_trees_regression": sklearn.ensemble.GradientBoostingRegressor,
-                    "gradient_boosting_trees_classification": sklearn.ensemble.GradientBoostingClassifier,
-                    "hist_gradient_boosting_regression": sklearn.ensemble.HistGradientBoostingRegressor,
-                    "hist_gradient_boosting_classification": sklearn.ensemble.HistGradientBoostingClassifier,
-                    "random_forest_regression": sklearn.ensemble.RandomForestRegressor,
-                    "random_forest_classification": sklearn.ensemble.RandomForestClassifier,
-                    "xgboost_regression": xgb.XGBRegressor,
-                    "xgboost_classification": xgb.XGBClassifier,
-                }[self.algorithm_name + "_" + self.project.objective](**self.hyperparams)
-                if len(self.snapshot.y_column_name) > 0:
+                algorithm = Model.algorithm_from_name_and_objective(self.algorithm_name, self.project.objective)
+                self._algorithm = algorithm(**self.hyperparams)
+                if len(self.snapshot.y_column_name) > 1:
                     if self.project.objective == "regression" and self.algorithm_name in [
                         "bayesian_ridge",
                         "automatic_relevance_determination",
@@ -627,12 +669,43 @@ class Model(object):
         """
         X_train, X_test, y_train, y_test = snapshot.data()
 
+        search_args = {"scoring": self.project.hyperparam_score_name, "error_score": "raise", **self.search_args}
+        if self.search == "grid":
+            self._algorithm = sklearn.model_selection.GridSearchCV(self.algorithm, self.search_params, **search_args)
+        elif self.search == "random":
+            self._algorithm = sklearn.model_selection.RandomizedSearchCV(
+                self.algorithm, self.search_params, **search_args
+            )
+        elif self.search is not None:
+            raise PgMLException(
+                f"Unknown hyperparam search `{self.search}`, available options are: ['grid', 'random']."
+            )
+
+        X_train, X_test, y_train, y_test = snapshot.data()
+
         # Train the model
-        self.algorithm.fit(X_train, y_train)
+        result = self.algorithm.fit(X_train, y_train)
+
+        metrics = {}
+        if self.search:
+            self._algorithm = result.best_estimator_
+            self.hyperparams = result.best_params_
+            for key, value in result.cv_results_.items():
+                if isinstance(value, numpy.ndarray):
+                    result.cv_results_[key] = value.tolist()
+            metrics["search_results"] = {
+                "best_index": int(result.best_index_),
+                "n_splits": int(result.n_splits_),
+                **result.cv_results_,
+            }
 
         # Test
         y_pred = self.algorithm.predict(X_test)
-        metrics = {}
+        if hasattr(self.algorithm, "predict_proba"):
+            y_prob = self.algorithm.predict_proba(X_test)
+            y_test = numpy.array(y_test).flatten()
+        else:
+            y_prob = None
         if self.project.objective == "regression":
             metrics["mean_squared_error"] = mean_squared_error(y_test, y_pred)
             metrics["r2"] = r2_score(y_test, y_pred)
@@ -640,6 +713,10 @@ class Model(object):
             metrics["f1"] = f1_score(y_test, y_pred, average="weighted")
             metrics["precision"] = precision_score(y_test, y_pred, average="weighted")
             metrics["recall"] = recall_score(y_test, y_pred, average="weighted")
+            metrics["accuracy"] = accuracy_score(y_test, y_pred)
+            if y_prob is not None:
+                metrics["log_loss"] = log_loss(y_test, y_prob)
+                metrics["roc_auc"] = roc_auc_score(y_test, y_prob, average="weighted", multi_class="ovo")
 
         # Save the model
         self.__dict__ = dict(
@@ -648,7 +725,8 @@ class Model(object):
             UPDATE pgml.models
             SET pickle = '\\x{pickle.dumps(self.algorithm).hex()}',
                 status = 'successful',
-                metrics = {q(json.dumps(metrics))},
+                hyperparams = {q(self.hyperparams)},
+                metrics = {q(metrics)},
                 updated_at = clock_timestamp()
             WHERE id = {q(self.id)}
             RETURNING *
@@ -678,10 +756,23 @@ class Model(object):
         """
         # TODO: add metrics for tracking prediction volume/accuracy by model
         # TODO: smarter treatment for images rather than flattening
-        y = self.algorithm.predict([flatten(data)])
-        if self.project.objective == "regression" and len(self.snapshot.y_column_name) == 1:
-            y = y[0]
-        return y
+        y = self.algorithm.predict([numpy.array(data).flatten()])
+        if isinstance(y[0], numpy.ndarray) and len(self.snapshot.y_column_name) == 1:
+            # HACK: it's unfortunate that some sklearn models always return a 2D array, and some only return a 2D array for multiple outputs.
+            return y[0][0]
+        return y[0]
+
+
+def snapshot(
+    relation_name: str,
+    y_column_name: str,
+    test_size: float or int = 0.25,
+    test_sampling: str = "random",
+):
+    """Create a snapshot of a relation.
+
+    Same args as train() below."""
+    return Snapshot.create(relation_name, y_column_name, test_size, test_sampling)
 
 
 def train(
@@ -691,7 +782,10 @@ def train(
     y_column_name: str = None,
     algorithm_name: str = "linear",
     hyperparams: dict = {},
-    test_size: float or int = 0.1,
+    search: str = None,
+    search_params: dict = {},
+    search_args: dict = {},
+    test_size: float or int = 0.25,
     test_sampling: str = "random",
 ):
     """Create a regression model from a table or view filled with training data.
@@ -702,10 +796,10 @@ def train(
         relation_name (str): the table or view that stores the training data
         y_column_name (str): the column in the training data that acts as the label
         algorithm_name (str, optional): the algorithm used to implement the objective. Defaults to "linear". Valid values are ["linear", "svm", "random_forest", "gradient_boosting"].
-        test_size (float or int, optional): If float, should be between 0.0 and 1.0 and represent the proportion of the dataset to include in the test split. If int, represents the absolute number of test samples. If None, the value is set to the complement of the train size. If train_size is also None, it will be set to 0.25.
+        test_size (float or int, optional): If float, should be between 0.0 and 1.0 and represent the proportion of the dataset to include in the test split. If int, represents the absolute number of test samples. If None, the value is set to 0.25.
         test_sampling: (str, optional): How to sample to create the test data. Defaults to "random". Valid values are ["first", "last", "random"].
     """
-
+    # Project
     try:
         project = Project.find_by_name(project_name)
         if objective is not None and objective != project.objective:
@@ -716,12 +810,10 @@ def train(
     except PgMLException:
         project = Project.create(project_name, objective)
 
-    if algorithm_name is None:
-        algorithm_name = "linear"
-
     if objective not in ["regression", "classification"]:
         raise PgMLException(f"Unknown objective `{objective}`, available options are: regression, classification.")
 
+    # Create or use an existing snapshot.
     if relation_name is None:
         snapshot = project.last_snapshot
         if snapshot is None:
@@ -732,13 +824,22 @@ def train(
             raise PgMLException(
                 f"You must pass a `relation_name` to use a different `y_column_name` than previous runs. {y_column_name} vs {snapshot.y_column_name}"
             )
-
     else:
         snapshot = Snapshot.create(relation_name, y_column_name, test_size, test_sampling)
 
-    model = Model.create(project, snapshot, algorithm_name, hyperparams)
+    # Model
+    if algorithm_name is None:
+        algorithm_name = "linear"
+
+    # Default repeatable random state when possible
+    algorithm = Model.algorithm_from_name_and_objective(algorithm_name, objective)
+    if "random_state" in algorithm().get_params() and "random_state" not in hyperparams:
+        hyperparams["random_state"] = 0
+
+    model = Model.create(project, snapshot, algorithm_name, hyperparams, search, search_params, search_args)
     model.fit(snapshot)
 
+    # Deployment
     if (
         project.deployed_model is None
         or (
